from typing import Callable
from kola.lib.debugger.base import BaseDebugger


class CommandDebugger(BaseDebugger):
    """
    output the command and additional arguments to stdout
    """
    def __getitem__(self, __key: str) -> Callable[..., None]:
        if __key != "@exception":
            return lambda *args, **kwds: \
                print(f"## [DEBUG] command {__key} with args {args} kwds {kwds}")
        return super().__getitem__(__key)



from abc import ABCMeta
from typing import (Any, Callable, Dict, Iterable, Optional, Set, Tuple, Union,
                    overload)

from typing_extensions import Self

from .command import Command, CommandLike
from .mask import Mask, ClassTypeMask


class CommandSetMeta(ABCMeta):
    __command_field__: Set[CommandLike]
    __virtual_table__: Dict[str, str]

    def __new__(cls, name: str, bases: Tuple[type, ...], attr: Dict[str, Any], **kwds: Any) -> Self: ...
    def generate_raw_commands(self) -> Dict[str, Any]: ...
    @overload
    def register_command(
        self, __func: Callable[..., Any], **kwds) -> Command: ...
    @overload
    def register_command(
        self, __name: Optional[str] = ..., *, alias: Union[Iterable[str], str] = ..., virtual: bool = False, **kwds: Any
    ) -> Callable[[Callable[..., Any]], Command]: ...
    @overload
    def register_text(
        self, __func: Callable[..., Any], **kwds) -> Command: ...
    @overload
    def register_text(
        self, *, alias: Union[Iterable[str], str] = ..., virtual: bool = False, **kwds: Any
    ) -> Callable[[Callable[..., Any]], Command]: ...
    @overload
    def register_number(
        self, __func: Callable[..., Any], **kwds) -> Command: ...
    @overload
    def register_number(
        self, *, alias: Union[Iterable[str], str] = ..., virtual: bool = False, **kwds: Any
    ) -> Callable[[Callable[..., Any]], Command]: ...
    @overload
    def register_annotation(
        self, __func: Callable[..., Any], **kwds) -> Command: ...
    @overload
    def register_annotation(
        self, *, alias: Union[Iterable[str], str] = ..., virtual: bool = False, **kwds: Any
    ) -> Callable[[Callable[..., Any]], Command]: ...
    

class CommandSet(metaclass=CommandSetMeta):
    __slots__ = ["raw_command_set", "_bound_command_cache"]

    raw_command_set: Dict[str, Callable]
    _bound_command_set: Dict[str, Callable]

    def __init__(self) -> None: ...
    def check_virtual(self, command: Command) -> bool: ...
    def get(self, __key: str, default: Optional[Callable] = ...) -> Optional[Callable]: ...
    @classmethod
    def mask(cls, type: Union["Mask.MType", str] = "") -> ClassTypeMask: ...
    def __kola_caller__(self, command: Command, args: tuple, kwargs: Dict[str, Any], **kwds) -> Any: ...
    def __getitem__(self, __key: str) -> Callable: ...



"""
Copyright 2022 Ovizro

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import sys
from argparse import ArgumentParser
from traceback import print_exc

from .lexer import BaseLexer, FileLexer, StringLexer
from .exception import KoiLangError
from .lib import load_library, main_class_from_module
from .lib.debugger import KoiLangRunner, CommandDebugger

from . import __version__


def _read_stdin() -> str:
    try:
        sys.stdout.write("$kola: ")
        sys.stdout.flush()
        s = sys.stdin.readline()
        while s.endswith("\\\n"):
            sys.stdout.write("$...: ")
            sys.stdout.flush()
            s += sys.stdin.readline()
    except KeyboardInterrupt:
        sys.exit()
    return s


if __name__ == "__main__":
    parser = ArgumentParser("kola")
    parser.add_argument("file", default=None, nargs="?")
    parser.add_argument("-i", "-c", "--inline", help="parse inline string")
    parser.add_argument("-s", "--script", help="parser script")
    parser.add_argument("-d", "--debug", help="dubugger type", choices=["token", "command"])
    parser.add_argument("--encoding", help="file encoding", default="utf-8")

    namespace = parser.parse_args()

    encoding = namespace.encoding
    if namespace.file:
        lexer = FileLexer(namespace.file, encoding=encoding)
    elif namespace.inline:
        lexer = StringLexer(namespace.inline, encoding=encoding)
    elif not sys.stdin.isatty():
        lexer = BaseLexer(encoding=encoding)
    else:
        lexer = None

    runner_type = "Runner"
    if namespace.debug == "token":
        runner_type = "Token Debugger"
        if lexer is None:
            print(f"KoiLang {runner_type} {__version__} on Python {sys.version}")
            lexer = BaseLexer()
        while True:
            try:
                for i in lexer:
                    print(i)
                break
            except KeyboardInterrupt:
                break
            except KoiLangError:
                print_exc()
    else:
        if namespace.debug == "command":
            command_cls = CommandDebugger
            runner_type = "Command Debugger"
        elif namespace.script:
            command_cls = main_class_from_module(load_library(namespace.script))
        else:
            command_cls = KoiLangRunner

        with command_cls().exec_block() as command_set:
            if lexer:
                command_set.parse(lexer)
            else:
                print(f"KoiLang {runner_type} {__version__} on Python {sys.version}")
                while True:
                    command_set.parse(_read_stdin())



from abc import ABCMeta
from types import MethodType
from typing import Any, Callable, Dict, Optional, Set, Tuple, Type, Union
from typing_extensions import Self

from .command import CommandLike, Command


class CommandSetMeta(ABCMeta):
    """
    metaclass for all command sets
    """
    __command_field__: Set[CommandLike]
    __virtual_table__: Dict[str, str]

    def __new__(cls, name: str, bases: Tuple[Type, ...], attr: Dict[str, Any], **kwds: Any) -> Self:
        command_field = set()
        virtual_table = attr.get("__virtual_table__", {})
        for i in bases:
            if isinstance(i, CommandSetMeta):
                virtual_table.update(i.__virtual_table__)
        for k, v in attr.items():
            if isinstance(v, Command):
                if v.virtual:
                    virtual_table[k] = v.__name__
                elif k in virtual_table:
                    v.virtual = True
            elif k in virtual_table:
                attr[k] = v = Command(virtual_table[k], v, virtual=True)
            elif not isinstance(v, CommandLike):
                continue
            command_field.add(v)
        attr["__command_field__"] = command_field
        attr["__virtual_table__"] = virtual_table
        return super().__new__(cls, name, bases, attr, **kwds)

    def generate_raw_commands(self) -> Dict[str, Any]:
        cmd_set = {}
        for cls in reversed(self.__mro__):
            cls: Type
            if not isinstance(cls, CommandSetMeta):
                continue
            for c in cls.__command_field__:
                cmd_set.update(c.__kola_command__())
        return cmd_set

    def _command_register_factory(cmd_name: Optional[str] = None):  # type: ignore
        def inner(
            self,
            __func_or_name: Union[Callable, str, None] = None,
            **kwds
        ) -> Union[Callable[[Callable], Command], Command]:
            def wrapper(wrapped_func: Callable[..., Any]) -> Command:
                if cmd_name is not None:
                    assert not isinstance(__func_or_name, str)
                    name = cmd_name
                elif isinstance(__func_or_name, str):
                    name = __func_or_name
                else:
                    name = wrapped_func.__name__
                cmd = Command(name, wrapped_func, **kwds)
                self.__command_field__.add(cmd)
                return cmd
            if callable(__func_or_name):
                return wrapper(__func_or_name)
            else:
                return wrapper
        return inner

    register_command = _command_register_factory()
    register_text = _command_register_factory("@text")
    register_number = _command_register_factory("@number")
    register_annotation = _command_register_factory("@annotation")

    del _command_register_factory
    
    def __repr__(self) -> str:
        return f"<kola command set '{self.__qualname__}'>"


class CommandSet(object, metaclass=CommandSetMeta):
    __slots__ = ["raw_command_set", "_bound_command_cache"]

    def __init__(self) -> None:
        super().__init__()
        self.raw_command_set = self.__class__.generate_raw_commands()
        self._bound_command_cache = {}

    def get(self, __key: str, default: Optional[Callable] = None) -> Optional[Callable]:
        """
        get command in the command set

        NOTE: This method will only try to get its own commands.
        If you want to get a command as a normal case, use `__getitem__` instead.
        """
        cache = self._bound_command_cache.get(__key, None)
        if cache is not None:
            return cache
        raw_cmd = self.raw_command_set.get(__key, default)
        if raw_cmd is default:
            return raw_cmd
        
        assert raw_cmd
        bound_cmd = MethodType(raw_cmd, self)
        self._bound_command_cache[__key] = bound_cmd
        return bound_cmd

    def check_virtual(self, command: Command) -> bool:
        return command is self.raw_command_set[command.__name__]

    @classmethod
    def mask(cls, type: Union["Mask.MType", str] = "") -> "Mask":
        return ClassTypeMask(cls, type)  # type: ignore

    def __kola_caller__(self, command: Command, args: tuple, kwargs: Dict[str, Any], **kwds: Any) -> Any:
        """hook function used to change the calling behavior of the `Command` class

        :param command: the `Command` object being invoked
        :type command: Command
        :param args: call positional arguments
        :type args: Tuple[Any]
        :param kwargs: call keyword arguments
        :type kwargs: Dict[str, Any]
        :return: command return value
        :rtype: Any
        """
        return command.__func__(self, *args, **kwargs)

    def __getitem__(self, __key: str) -> Callable:
        cmd = self.get(__key)
        if cmd is None:
            raise KeyError(f"unknown command '{__key}'")
        return cmd
    
    def __repr__(self) -> str:
        return f"<kola {self.__class__.__name__} object at 0x{id(self):08X}>"


from .mask import Mask, ClassTypeMask



"""
Kola sub libraries support modules

Include some useful builtin kola sub module.
And modules imported from `load_library()` will be mounted as a sub module.
"""

import inspect
import os
import re
import sys
from glob import glob
from importlib.util import spec_from_file_location
from importlib._bootstrap import _load
from types import ModuleType
from typing import List, NamedTuple, Optional, Type

from ..klvm import CommandSetMeta, KoiLang


if "KOLA_LIB_PATH" in os.environ:  # pragma: no cover
    KOLA_LIB_PATH = os.environ["KOLA_LIB_PATH"].split(os.pathsep)
else:
    KOLA_LIB_PATH = ['.', './kola_lib', os.path.dirname(__file__)]


_module_pattern = re.compile(r"^[A-Za-z_]\w*(?:\.[A-Za-z_]\w*)*$", re.A)


class KolaSpec(NamedTuple):
    """
    Kola module spec
    
    Usage:
        __kola_spec__ = KolaSpec(
            __name__,
            collect_all()
        )
    """
    name: str
    command_set_classes: List[CommandSetMeta]
    
    doc: Optional[str] = None
    main_class: Optional[Type[KoiLang]] = None


def collect_all() -> List[CommandSetMeta]:
    """get all CommandSet class defined in the module

    :return: a list of CommandSet class found
    :rtype: List[CommandSetMeta]
    """
    frame = inspect.currentframe()
    if not frame or not frame.f_back:  # pragma: no cover
        raise ValueError("cannot read variables from the frame")
    return [
        i for i in frame.f_back.f_locals.values()
        if isinstance(i, CommandSetMeta)
    ]


def load_library(name: str, paths: List[str] = KOLA_LIB_PATH) -> ModuleType:
    """load a Phython script or package as a Kola module

    :param name: module name like 'mymodule.sub'
    :type name: str
    :param paths: base path to load the module, defaults to KOLA_LIB_PATH
    :type paths: List[str], optional
    :raises ValueError: illegal module name
    :raises ImportError: fail to load the module
    :return: the Kola module
    :rtype: ModuleType
    """
    if not _module_pattern.match(name):  # pragma: no cover
        raise ValueError(f"illegal module name '{name}'")
    module_name = "kola.lib." + name
    if module_name in sys.modules:
        return sys.modules[module_name]

    for b in paths:
        spec = None
        b = os.path.abspath(b)
        full_path = os.path.join(b, name)
        if os.path.isdir(full_path):
            spec = spec_from_file_location(
                module_name, os.path.join(full_path, "__init__.py"),
                submodule_search_locations=[full_path]
            )
        else:
            for i in glob(full_path + '.*'):
                spec = spec_from_file_location(module_name, i)
                if spec:
                    full_path = i
                    break
        if spec:
            spec.name = module_name
            return _load(spec)
    raise ImportError(f"cannot load script {name}", name=name)  # pragma: no cover


def main_class_from_module(module: ModuleType) -> Type[KoiLang]:
    spec = getattr(module, "__kola_spec__", None)
    if spec:
        if not isinstance(spec, KolaSpec):
            spec = KolaSpec(*spec)
        if spec.main_class:
            return spec.main_class
        for i in spec.command_set_classes:
            if issubclass(i, KoiLang):
                return i
    else:
        for i in module.__dict__.values():
            if isinstance(i, type) and issubclass(i, KoiLang) and i is not KoiLang:
                return i
    raise ValueError(  # pragma: no cover
        f"no available main class for kola module '{spec.name if spec else module.__name__}'"
    )



import os
from inspect import signature, Signature
from functools import lru_cache
from typing import Any, Callable, Dict, Optional, Tuple, Type, Union
from typing_extensions import Self

from ..writer import BaseWriter, FileWriter, StringWriter
from .command import Command
from .environment import Environment
from .handler import AbstractHandler
from .koilang import KoiLang


@lru_cache()
def _default_writer_factory(name: str, sig: Optional[Signature] = None):
    if name == "@text":
        return BaseWriter.write_text
    elif name == "@annotation":
        return BaseWriter.write_annotation

    def inner(writer: BaseWriter, *args, **kwds) -> None:
        if sig:
            # check writer arguments
            sig.bind(writer, *args, **kwds)
        if name != "@number":
            args = (name,) + args
        writer.write_command(*args, **kwds)
    return inner


class KoiLangWriter(KoiLang):
    """
    writer class for the KoiLang class

    The class should not be used directly.
    The usual way to use it is to call YourKoiLang.writer().
    """
    def __init__(self, ___writer: Union[str, bytes, os.PathLike, BaseWriter, None] = None) -> None:
        super().__init__()
        if ___writer is None:
            self._writer = StringWriter(
                command_threshold=self.__class__.__command_threshold__
            )
        elif isinstance(___writer, BaseWriter):
            self._writer = ___writer
        else:
            self._writer = FileWriter(
                ___writer,
                command_threshold=self.__class__.__command_threshold__,
                encoding=self.__class__.__text_encoding__
            )
    
    def push_apply(self, __push_cache: Environment) -> None:
        self._writer.inc_indent()
        return super().push_apply(__push_cache)
    
    def pop_prepare(self, __env_type: Optional[Type[Environment]] = None) -> Environment:
        self._writer.dec_indent()
        return super().pop_prepare(__env_type)
    
    def newline(self) -> None:
        self._writer.newline()
    
    def getvalue(self) -> str:
        if not isinstance(self._writer, StringWriter):  # pragma: no cover
            raise TypeError("only `StringWriter` object can use 'getvalue' method")
        return self._writer.getvalue()

    def __enter__(self) -> Self:
        self._writer.__enter__()
        return self

    def __exit__(self, *args) -> None:
        return self._writer.__exit__(*args)


@KoiLangWriter.register_handler
class WriterHandler(AbstractHandler):
    __slots__ = []

    priority = 3
    owner: "KoiLangWriter"

    def __call__(
        self,
        command: Command,
        args: Tuple,
        kwargs: Dict[str, Any],
        *,
        writer_func: Optional[Callable] = None,
        **kwds: Any
    ) -> Any:
        name = command.__name__
        if name.startswith('@') and name not in ["@text", "@number", "@annotation"]:  # pragma: no cover
            return super().__call__(command, args, kwargs, **kwds)
        
        if not writer_func:
            writer_func = _default_writer_factory(
                command.__name__, signature(command.__func__)
            )
        return writer_func(self.owner._writer, *args, **kwargs)



import os
import sys
from contextlib import contextmanager, suppress
from functools import partial
from threading import Lock
from types import TracebackType, new_class
from typing import (Any, Callable, Dict, Generator, List, Optional, Tuple,
                    Type, TypeVar, Union, overload)
from typing_extensions import Literal, Self

from ..exception import KoiLangError
from ..lexer import BaseLexer, FileLexer, StringLexer
from ..parser import Parser
from .command import Command
from .commandset import CommandSet, CommandSetMeta
from .environment import Environment


_T_EnvCls = TypeVar("_T_EnvCls", bound=Type[Environment])
_T_Handler = TypeVar("_T_Handler", bound=Type["AbstractHandler"])


class KoiLangMeta(CommandSetMeta):
    """
    metaclass for KoiLang class

    Provide encoding and command threshold support.
    """
    __text_encoding__: str
    __text_lstrip__: bool
    __command_threshold__: int
    __command_handlers__: List[Type["AbstractHandler"]]

    def __new__(
        cls,
        name: str,
        bases: Tuple[type, ...],
        attr: Dict[str, Any],
        command_threshold: int = 0,
        encoding: Optional[str] = None,
        lstrip_text: Optional[bool] = None,
        **kwds: Any
    ) -> Self:
        """
        create a top-level language class

        :param name: class name
        :type name: str
        :param bases: class bases
        :type bases: Tuple[type, ...]
        :param attr: class namespace
        :type attr: Dict[str, Any]
        :param command_threshold: the `#` prefix length of commands, defaults to 0
        :type command_threshold: int, optional
        :param encoding: encoding for file parsing, defaults to None
        :type encoding: Optional[str], optional
        :param lstrip_text: whether to remove text indentation, defaults to True
        :type lstrip_text: bool, optional
        :return: new class
        :rtype: KoiLangMeta
        """
        # if not base KoiLang class, set a default value
        if not any(isinstance(i, cls) for i in bases):
            command_threshold = command_threshold or 1
            encoding = encoding or "utf-8"
            lstrip_text = lstrip_text if lstrip_text is not None else True
            if "__command_handlers__" not in attr:
                attr["__command_handlers__"] = []
        
        if command_threshold:
            assert command_threshold >= 0
            attr["__command_threshold__"] = command_threshold
        if lstrip_text is not None:
            attr["__text_lstrip__"] = lstrip_text
        if encoding:
            attr["__text_encoding__"] = encoding
        return super().__new__(cls, name, bases, attr, **kwds)

    def register_environment(self, env_class: _T_EnvCls) -> _T_EnvCls:
        self.__command_field__.add(env_class)
        return env_class
    
    def register_handler(self, handler: _T_Handler) -> _T_Handler:
        if "__command_handlers__" not in self.__dict__:
            # copy the handler list to avoid changing the base classes
            self.__command_handlers__ = self.__command_handlers__.copy()
        self.__command_handlers__.append(handler)
        return handler
    
    @property
    def writer(self) -> Type:
        """writer class for KoiLang file building

        :return: the writer class, which is a subclass of KoiLangWriter and current KoiLang class
        :rtype: Type[KoiLang, Self]
        """
        # sourcery skip: inline-immediately-returned-variable
        cache = None
        with suppress(AttributeError):
            cache = self.__writer_class
            return cache
        cache = new_class(f"{self.__qualname__}.writer", (KoiLangWriter, self))
        self.__writer_class = cache
        return cache


class KoiLang(CommandSet, metaclass=KoiLangMeta):
    """
    main class for KoiLang virtual machine

    `KoiLang` class is the top-level interface of 'kola' package.
    Just create a subclass to define your own markup language based on KoiLang.
    """
    __slots__ = ["_handler", "_lock", "__top", "__exec_level"]

    def __init__(self) -> None:
        super().__init__()
        self._lock = Lock()
        self.__top = self
        self.__exec_level = 0
        self._handler = build_handlers(self.__class__.__command_handlers__, self)
    
    def push_prepare(self, __env_type: Union[Type[Environment], Environment]) -> Environment:
        if not isinstance(__env_type, Environment):
            __env_type = __env_type(self.__top)
        __env_type.set_up(self.__top)
        return __env_type

    def push_apply(self, __env_cache: Environment) -> None:
        assert __env_cache.back is self.__top
        with self._lock:
            self.__top = __env_cache
    
    def pop_prepare(self, __env_type: Optional[Type[Environment]] = None) -> Environment:
        top = self.__top
        if top is self:  # pragma: no cover
            raise ValueError('cannot pop the inital environment')
        if __env_type is None:
            assert isinstance(top, Environment)
        else:
            while isinstance(top, Environment) and top.__class__.__env_autopop__:
                if isinstance(top, __env_type):
                    break
                top = top.back
            else:
                if not isinstance(top, __env_type):  # pragma: no cover
                    raise ValueError("unmatched environment")
        return top

    def pop_apply(self, __env_cache: Environment) -> None:
        with self._lock:
            top = self.__top
            self.__top = __env_cache.back
        while isinstance(top, Environment):
            top.tear_down(self.__top)
            if top is __env_cache:
                break
            top = top.back
        else:
            raise ValueError('cannot pop the inital environment')
    
    def add_handler(self, handler: Union[Type["AbstractHandler"], "AbstractHandler"]) -> "AbstractHandler":
        if isinstance(handler, type):
            handler = handler(self)
        self._handler = self._handler.insert(handler)
        return handler
    
    def remove_handler(self, handler: "AbstractHandler") -> None:
        hdl = self._handler.remove(handler)
        if hdl is None:  # pragma: no cover
            raise ValueError("cannot remove all handlers")
        self._handler = hdl

    def __parse(self, __lexer: BaseLexer, *, close_lexer: bool = True) -> None:
        parser = Parser(__lexer, self)
        try:
            with self.exec_block():
                while True:
                    try:
                        # Parser.exec() is a fast C level loop
                        parser.exec()
                    except KoiLangError:
                        if not self.on_exception(*sys.exc_info()):
                            raise
                    else:
                        break
        finally:
            if close_lexer:
                __lexer.close()
    
    def __parse_and_ret(self, __lexer: BaseLexer, *, close_lexer: bool = True) -> Generator[Any, None, None]:
        parser = Parser(__lexer, self)
        try:
            with self.exec_block():
                while True:
                    try:
                        yield from parser
                    except KoiLangError:
                        if not self.on_exception(*sys.exc_info()):
                            raise
                    else:
                        break
        finally:
            if close_lexer:
                __lexer.close()

    @overload
    def parse(self, lexer: Union[BaseLexer, str], *, with_ret: Literal[False] = False, close_lexer: bool = True) -> None: ...
    @overload  # noqa: E301
    def parse(
        self,
        lexer: Union[BaseLexer, str],
        *,
        with_ret: Literal[True],
        close_lexer: bool = True
    ) -> Generator[Any, None, None]: ...
    
    def parse(self, lexer: Union[BaseLexer, str], *, with_ret: bool = False, close_lexer: bool = True) -> Any:
        """parse kola text

        :param lexer: Lexer object or legal KoiLang string
        :type lexer: Union[BaseLexer, str]
        :param with_ret: if true, return a gererater where command returns would be yielded, defaults to False
        :type with_ret: bool, optional
        :param close_lexer: whether or not to close the lexer, defaults to True
        :type close_lexer: bool, optional
        :raises ValueError: when a KoiLang string given without trying to close it
        :return: return a generator if `with_ret` set
        :rtype: Generator[Any, None, None] or None
        """
        if isinstance(lexer, str):
            if not close_lexer:  # pragma: no cover
                raise ValueError("inner string lexer must be closed at the end of parsing")
            lexer = StringLexer(
                lexer,
                encoding=self.__class__.__text_encoding__,
                command_threshold=self.__class__.__command_threshold__,
                no_lstrip=not self.__class__.__text_lstrip__
            )
        if with_ret:
            return self.__parse_and_ret(lexer, close_lexer=close_lexer)
        else:
            self.__parse(lexer, close_lexer=close_lexer)
        
    def parse_file(self, path: Union[str, bytes, os.PathLike], *, encoding: Optional[str] = None, **kwds: Any) -> Any:
        """
        parse a kola file.
        """
        return self.parse(
            FileLexer(
                path, encoding=encoding or self.__class__.__text_encoding__,
                command_threshold=self.__class__.__command_threshold__,
                no_lstrip=not self.__class__.__text_lstrip__
            ), **kwds
        )
    
    @contextmanager
    def exec_block(self) -> Generator[Self, None, None]:
        if not self.__exec_level:
            self.at_start()
        with self._lock:
            self.__exec_level += 1
        try:
            yield self
        finally:
            with self._lock:
                self.__exec_level -= 1
            if not self.__exec_level:
                self.at_end()

    def __getitem__(self, __key: str) -> Callable:
        if self.__top is self:
            return super().__getitem__(__key)
        return self.__top[__key]

    def __kola_caller__(self, command: Command, args: tuple, kwargs: Dict[str, Any], **kwds: Any) -> Any:
        return self._handler(command, args, kwargs, **kwds)

    @property
    def top(self) -> CommandSet:
        return self.__top
    
    @property
    def home(self) -> Self:
        return self
    
    @property
    def handlers(self) -> "HandlerSequence":
        return HandlerSequence(self._handler)

    @partial(Command, "@start", virtual=True)
    def at_start(self) -> None:
        """
        parser initalize command
        
        It is called before parsing start.
        """

    @partial(Command, "@end", virtual=True)
    def at_end(self) -> None:
        """
        parser finalize command
        
        It is called after parsing end. And the return value
        will be that of 'parse' method.
        """
    
    @partial(Command, "@exception", virtual=True)
    def on_exception(self, exc_type: Type[KoiLangError], exc_ins: Optional[KoiLangError], traceback: TracebackType) -> None:
        """
        exception handling command

        It is called when a KoiLang error occurs.
        If the command wishes to suppress the exception, it should a true value.
        """


from .handler import AbstractHandler, HandlerSequence, build_handlers
from .writer import KoiLangWriter



from types import new_class
from typing import Any, Callable, Optional, Type, Union

from .command import Command
from .commandset import CommandSet
from .environment import Environment, EnvironmentEntry, EnvironmentExit
from .koilang import KoiLang


def _kola_decorator_factory(cmd_name: Optional[str] = None, cmd_type: Type[Command] = Command):
    def kola_decorator(
        func: Union[Callable[..., Any], str, None] = None,
        **kwds
    ) -> Union[Command, Callable[[Callable], Command]]:
        def wrapper(wrapped_func: Callable[..., Any]) -> Command:
            nonlocal cmd_type
            if cmd_name is not None:
                assert not isinstance(func, str)
                name = cmd_name
            elif isinstance(func, str):
                name = func
            else:
                name = wrapped_func.__name__
            return cmd_type(name, wrapped_func, **kwds)
        if callable(func):
            return wrapper(func)
        else:
            return wrapper
    return kola_decorator


kola_command = _kola_decorator_factory()
kola_text = _kola_decorator_factory("@text")
kola_number = _kola_decorator_factory("@number")
kola_annotation = _kola_decorator_factory("@annotation")
kola_env_enter = _kola_decorator_factory(cmd_type=EnvironmentEntry)
kola_env_exit = _kola_decorator_factory(cmd_type=EnvironmentExit)


def _kola_class_decorator_factory(base: Type[CommandSet] = CommandSet):
    def kola_class_decoractor(kola_cls: Union[Type[KoiLang], str, None] = None, **kwds):
        def wrapper(wrapped_class: Type) -> Type[CommandSet]:
            if isinstance(kola_cls, str):
                env_name = kola_cls
            else:
                env_name = wrapped_class.__name__
            
            return new_class(
                env_name,
                (base,),
                kwds,
                lambda attrs: attrs.update(wrapped_class.__dict__)
            )
        if isinstance(kola_cls, type):
            return wrapper(kola_cls)
        else:
            return wrapper
    return kola_class_decoractor


kola_command_set = _kola_class_decorator_factory()
kola_environment = _kola_class_decorator_factory(Environment)
kola_main = _kola_class_decorator_factory(KoiLang)



import os
from typing import Any, ClassVar, Final, FrozenSet, Union, final, TypedDict
from typing_extensions import Self


S_CMD: int
S_CMD_N: int
S_TEXT: int
S_LITERAL: int
S_STRING: int
S_NUM: int
S_NUM_H: int
S_NUM_B: int
S_NUM_F: int
S_CLN: int
S_CMA: int
S_SLP: int
S_SRP: int
S_ANNOTATION: int
F_DISABLED: int
F_LSTRIP_TEXT: int


class _LexerData(TypedDict):
    filename: str
    encoding: str
    command_threshold: int
    flag: int


@final
class Token:
    """
    Token used in lexer

    Don't instantiate this class directly unless you make
    sure enough arguments provided.
    """

    syn: Final[int]
    val: Final[Any]

    def __new__(cls, syn: int, val: Any = ..., *, lineno: int = ..., raw_val: bytes = ...) -> Self: ...
    def get_flag(self) -> int: ...


class LexerConfig:
    """
    Python-level interface to access extra lexer data
    """
    data_names: ClassVar[FrozenSet[str]]

    def __init__(self, lexer: BaseLexer) -> None: ...
    def dict(self) -> _LexerData: ...
    def set(
        self,
        *,
        encoding: str = ...,
        command_threshold: int = ...,
        flag: int = ...,
        disabled: bool = ...,
        no_lstrip: bool = ...
    ) -> None: ...
    @property
    def filename(self) -> str: ...
    @property
    def encoding(self) -> str: ...
    @encoding.setter
    def encoding(self, val: str) -> None: ...
    @property
    def command_threshold(self) -> int: ...
    @command_threshold.setter
    def command_threshold(self, cmd_threshold: int) -> None: ...    
    @property
    def flag(self) -> int: ...
    @flag.setter
    def flag(self, val: int) -> None: ...
    @property
    def disabled(self) -> bool: ...
    @disabled.setter
    def disabled(self, val: bool) -> None: ...
    @property
    def no_lstrip(self) -> bool: ...
    @no_lstrip.setter
    def no_lstrip(self, val: bool) -> None: ...


class BaseLexer:
    def __init__(self, *, encoding: str = ..., command_threshold: int = 1, no_lstrip: bool = ...) -> None: ...
    def close(self) -> None: ...
    @property
    def filename(self) -> str: ...
    @property
    def lineno(self) -> int: ...
    @property
    def column(self) -> int: ...
    @property
    def config(self) -> LexerConfig: ...
    @property
    def closed(self) -> bool: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> Token: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *args) -> None: ...


class FileLexer(BaseLexer):
    def __init__(self, __path: Union[str, bytes, os.PathLike], *, encoding: str = ...,
                 command_threshold: int = 1, no_lstrip: bool = ...) -> None: ...
    @property
    def filename(self) -> Union[str, bytes, os.PathLike]: ...


class StringLexer(BaseLexer):
    content: Final[bytes]

    def __init__(self, content: Union[str, bytes], *, encoding: str = ...,
                 command_threshold: int = 1, no_lstrip: bool = ...) -> None: ...



cimport cython
from libc.stdio cimport stdin, FILE, fopen, fclose, EOF
from cpython cimport PyObject

from ._cutil cimport *


cdef extern from *:
    bint PyDescr_IsData(object desc)


cdef class Token:
    cdef:
        Token next     # used in grammar parser
    cdef readonly:
        TokenSyn syn
        object val
        bytes raw_val
        int lineno
    
    cpdef int get_flag(self)


cdef class LexerConfig:
    cdef LexerData* lexer_data
    cdef readonly BaseLexer lexer


@cython.no_gc
cdef class BaseLexer:
    cdef:
        yyscan_t scanner
        LexerData lexer_data
    cdef readonly:
        str encoding

    cpdef void close(self)
    cdef void set_error(self, const char* text) except *
    cdef (int, const char*, Py_ssize_t) next_syn(self) nogil
    cdef Token next_token(self)


cdef class FileLexer(BaseLexer):
    cdef:
        object _filenameo
        bytes _filenameb
        FILE* fp
    
    cpdef void close(self)


@cython.no_gc
cdef class StringLexer(BaseLexer):
    cdef readonly bytes content


from typing import Any, Callable, Iterable, Optional, Type, overload, Union

from .command import Command
from .commandset import CommandSet
from .environment import Environment, EnvironmentEntry, EnvironmentExit
from .koilang import KoiLang


@overload
def kola_command(func: Callable[..., Any], **kwds: Any) -> Command: ...
@overload
def kola_command(
    func: Optional[str] = None,
    *,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...,
    virtual: bool = False,
    **kwds: Any
) -> Callable[[Callable[..., Any]], Command]: ...
@overload
def kola_text(func: Callable[..., Any], **kwds: Any) -> Command: ...
@overload
def kola_text(
    *,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...,
    virtual: bool = False,
    **kwds: Any
) -> Callable[[Callable[..., Any]], Command]: ...
@overload
def kola_number(func: Callable[..., Any], **kwds: Any) -> Command: ...
@overload
def kola_number(
    *,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...,
    virtual: bool = False,
    **kwds: Any
) -> Callable[[Callable[..., Any]], Command]: ...
@overload
def kola_annotation(func: Callable[..., Any], **kwds: Any) -> Command: ...
@overload
def kola_annotation(
    *,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...,
    virtual: bool = False,
    **kwds: Any
) -> Callable[[Callable[..., Any]], Command]: ...
@overload
def kola_env_enter(func: Callable[..., Any], **kwds: Any) -> EnvironmentEntry: ...
@overload
def kola_env_enter(
    func: Optional[str] = None,
    *,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...,
    virtual: bool = False,
    **kwds: Any
) -> Callable[[Callable[..., Any]], EnvironmentEntry]: ...
@overload
def kola_env_exit(func: Callable[..., Any], **kwds: Any) -> EnvironmentExit: ...
@overload
def kola_env_exit(
    func: Optional[str] = None,
    *,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...,
    virtual: bool = False,
    **kwds: Any
) -> Callable[[Callable[..., Any]], EnvironmentExit]: ...
@overload
def kola_command_set(kola_cls: Type, **kwds: Any) -> Type[CommandSet]: ...
@overload
def kola_command_set(kola_cls: Optional[str] = None, **kwds: Any) -> Callable[[Type], Type[CommandSet]]: ...
@overload
def kola_environment(kola_cls: Type, **kwds: Any) -> Type[Environment]: ...
@overload
def kola_environment(kola_cls: Optional[str] = None, **kwds: Any) -> Callable[[Type], Type[Environment]]: ...
@overload
def kola_main(kola_cls: Type, **kwds: Any) -> Type[KoiLang]: ...
@overload
def kola_main(kola_cls: Optional[str] = None, **kwds: Any) -> Callable[[Type], Type[KoiLang]]: ...



from abc import ABC, abstractmethod
from operator import index as op_index
from threading import Lock
from typing import Any, ClassVar, Dict, Generator, Iterable, List, NamedTuple, Optional, SupportsIndex, Tuple, Type, TypeVar, Union
from typing_extensions import Self

from .mask import ClassNameMask, Mask
from .command import Command
from .commandset import CommandSet


_T_Handler = TypeVar("_T_Handler", bound="AbstractHandler")


class AbstractHandler(ABC):
    __slots__ = ["next", "owner"]

    priority: ClassVar[int] = 0

    def __init__(self, owner: "KoiLang", next: Optional["AbstractHandler"] = None) -> None:
        super().__init__()
        self.owner = owner
        self.next = next
    
    def insert(self, handler: _T_Handler) -> Union[Self, _T_Handler]:
        assert handler.next is None
        if self.priority <= handler.priority:
            handler.next = self
            return handler
        elif self.next is None:
            self.next = handler
        else:
            self.next = self.next.insert(handler)
        return self
    
    def remove(self, handler: _T_Handler) -> Union[Self, _T_Handler, None]:
        if self is handler:
            return self.next
        elif self.next:
            self.next = self.next.remove(handler)
            return self
        raise ValueError(f"{handler} is not in the handler chain")

    @abstractmethod
    def __call__(self, command: Command, args: Tuple, kwargs: Dict[str, Any], **kwds: Any) -> Any:
        if self.next is None:
            return
        return self.next(command, args, kwargs, **kwds)


class HandlerSequence:
    __slots__ = ["head"]

    def __init__(self, head: AbstractHandler) -> None:
        self.head = head
    
    def __getitem__(self, __index: SupportsIndex) -> AbstractHandler:
        hdl = self.__get(__index)
        if hdl is None:
            raise KeyError(__index)
        return hdl
    
    def __iter__(self) -> Generator[AbstractHandler, None, None]:
        head = self.head
        while head is not None:
            yield head
            head = head.next
    
    def __contains__(self, hdl: AbstractHandler) -> bool:
        if not isinstance(hdl, AbstractHandler):
            return NotImplemented
        return any(hdl is i for i in self)
    
    def __get(self, __index: SupportsIndex, default: Optional[AbstractHandler] = None) -> Optional[AbstractHandler]:
        i = op_index(__index)
        if i < 0:
            return list(self)[i]
        for hdl in self:
            if i == 0:
                return hdl
            i -= 1
        return default
    
    get = __get


def build_handlers(handlers: List[Type[AbstractHandler]], ins: Any) -> AbstractHandler:
    assert handlers
    base = handlers[0](ins)
    for i in handlers[1:]:
        hdl = i(ins)
        base = base.insert(hdl)
    return base


from .koilang import KoiLang


default_handler = KoiLang.register_handler


@default_handler
class CallerHandler(AbstractHandler):
    __slots__ = []

    def __call__(
        self,
        command: Command,
        args: Tuple,
        kwargs: Dict[str, Any],
        *,
        bound_instance: Optional[CommandSet] = None,
        **kwds: Any
    ) -> Any:
        ret = command.__func__(bound_instance or self.owner, *args, **kwargs)
        super().__call__(command, args, kwargs, bound_instance=bound_instance, ret_value=ret, **kwds)
        return ret


@default_handler
class SkipHandler(AbstractHandler):
    __slots__ = []

    priority = 10

    def __call__(
        self,
        command: Command,
        args: Tuple,
        kwargs: Dict[str, Any],
        *,
        skip: bool = False,
        **kwds: Any
    ) -> Any:
        if not skip:
            return super().__call__(command, args, kwargs, **kwds)


class _EnvChecker(NamedTuple):
    reachable: List[CommandSet]
    contains: List[CommandSet]

    def eval_masks(self, checker: CommandSet, names: Iterable[Union[str, Mask]]) -> List[Mask]:
        if isinstance(names, (str, Mask)):
            names = (names,)
        var_dict = {str(i): self.contains[i] for i in range(len(self.contains) - 1, -1, -1)}
        var_dict["top"] = self.contains[0]
        var_dict.update({'?': checker, "cur": checker, "current": checker})
        var_dict.update(base=self.contains[-1], __init__=self.contains[-1])
        return [i if isinstance(i, Mask) else ClassNameMask(i, **var_dict) for i in names]
    
    def ensure_envs(self, masks: List[Mask]) -> None:
        ng, pt = [], []
        for i in masks:
            if i.not_:
                ng.append(i)
            else:
                pt.append(i)

        if ((not ng or all(not self.check_mask(i) for i in ng)) and
                (not pt or any(self.check_mask(i) for i in pt))):
            return
        raise ValueError(f"unmatched environment name {self.reachable[0]}")  # pragma: no cover
    
    def check_mask(self, mask: Mask) -> bool:
        if mask.type == Mask.MType.default:
            env_set = self.reachable
        elif mask.type == Mask.MType.all:
            env_set = self.contains
        else:
            return self.reachable[0] in mask
        return any(i in mask for i in env_set)


@default_handler
class EnsureEnvHandler(AbstractHandler):
    __slots__ = []

    priority = 5

    def _eval_checker(self) -> _EnvChecker:
        reachable: List[CommandSet] = []
        top = self.owner.top
        while isinstance(top, Environment):
            reachable.append(top)
            if not top.__class__.__env_autopop__:
                contains = reachable.copy()
                while isinstance(top, Environment):
                    contains.append(top)
                    top = top.back
                contains.append(self.owner)
                break
            top = top.back
        else:
            # the base KoiLang object name is '__init__'
            reachable.append(self.owner)
            contains = reachable
        return _EnvChecker(reachable, contains)
    
    def __call__(
        self,
        command: Command,
        args: Tuple,
        kwargs: Dict[str, Any],
        *,
        bound_instance: Optional[CommandSet] = None,
        envs: Tuple[str, ...] = (),
        **kwds: Any
    ) -> Any:
        if envs:
            checker = self._eval_checker()
            masks = checker.eval_masks(bound_instance or self.owner, envs)
            checker.ensure_envs(masks)
        return super().__call__(command, args, kwargs, bound_instance=bound_instance, **kwds)


from .environment import Environment



import os
from pathlib import Path
from typing import ClassVar, Optional, Type, Union
from typing_extensions import Literal

from kola.klvm import CommandSet, Environment, KoiLang, kola_command, kola_text, kola_env_enter, kola_env_exit
from kola.lib import _module_pattern


class FastFiles(KoiLang):
    __slots__ = ["base_path", "__base_path"]

    def __init__(self, base: Union[str, os.PathLike, None] = None) -> None:
        super().__init__()
        if base is None:
            self.__base_path = Path.cwd()
        else:
            self.__base_path = Path(base)

    class FileContains(Environment):
        __slots__ = ["fp"]

        home: "FastFiles"
        Space: ClassVar[Type["FastFiles.Space"]]

        @kola_env_enter("file", envs="!$?")
        def open(self, path: Union[str, os.PathLike], *, encoding: str = "utf-8") -> None:
            base = self.home.base_path
            self.fp = (base / path).open("w", encoding=encoding)
        
        @kola_command
        def seek(self, __cookie: int, __whence: Union[Literal["SET", "CUR", "END"], int]) -> None:
            if __whence == "SET":
                __whence = os.SEEK_SET
            elif __whence == "CUR":
                __whence = os.SEEK_CUR
            elif __whence == "END":
                __whence = os.SEEK_END
            elif not isinstance(__whence, int):
                raise ValueError(f"invalid seek whence {__whence}")
            self.fp.seek(__cookie, __whence)
        
        @kola_text
        def text(self, text: str) -> None:
            self.fp.write(text)
        
        @property
        def path(self) -> str:
            return self.fp.name

        def tear_down(self, top: CommandSet) -> None:
            self.fp.close()
            return super().tear_down(top)
    
    class Space(Environment):
        __slots__ = ["parent_path", "path"]

        home: "FastFiles"
        FileContains: ClassVar[Type["FastFiles.FileContains"]]

        @kola_env_enter("space")
        def enter(self, name: Optional[str] = None, path: Optional[Union[str, os.PathLike]] = None) -> None:
            base_path = self.home.base_path
            self.parent_path = base_path
            if not path:
                assert name
                if not _module_pattern.match(name):
                    raise ValueError(f"invalid space name {name}")
                path = os.path.join(*name.split('.'))
            self.path = base_path / path
            self.path.mkdir(755, exist_ok=True)
            self.home.base_path = self.path
        
        @kola_env_exit("endspace")
        def exit(self) -> None:
            self.home.base_path = self.parent_path
        
        def set_up(self, top: CommandSet) -> None:
            if isinstance(top, FastFiles.FileContains):
                home = self.home
                cache = home.pop_prepare()
                home.pop_apply(cache)
                # update self.back
                self.back = home.top
            return super().set_up(top)
    
    def at_start(self) -> None:
        self.base_path = self.__base_path
    
    def at_end(self) -> None:
        del self.base_path


FastFiles.FileContains.Space = FastFiles.Space
FastFiles.Space.FileContains = FastFiles.FileContains



from libc.stdio cimport FILE
from ._cutil cimport yyscan_t, LexerData

cdef extern from "lex.yy.c":
    ctypedef LexerData* YY_EXTRA_TYPE
    ctypedef void* YY_BUFFER_STATE

    # main parser
    int yylex(yyscan_t yyscanner) nogil
    
    int yylex_init(yyscan_t* ptr_yy_globals) nogil
    int yylex_init_extra(YY_EXTRA_TYPE yy_user_defined, yyscan_t* ptr_yy_globals) nogil
    int yylex_destroy(yyscan_t yyscanner) nogil
    void yyrestart(FILE* input_file, yyscan_t yyscanner) nogil
    bint yylex_check(yyscan_t yyscanner) nogil

    # Accessor  methods (get/set functions) to struct members.
    int yyget_lineno(yyscan_t yyscanner) nogil
    int yyget_column(yyscan_t yyscanner) nogil
    FILE *yyget_in(yyscan_t yyscanner) nogil
    FILE *yyget_out(yyscan_t yyscanner) nogil
    int yyget_leng(yyscan_t yyscanner) nogil
    char* yyget_text(yyscan_t yyscanner) nogil
    void yyset_lineno(int _line_number, yyscan_t yyscanner) nogil
    void yyset_column(int _column_no , yyscan_t yyscanner) nogil
    void yyset_in(FILE* _in_str , yyscan_t yyscanner) nogil
    void yyset_out(FILE* _out_str , yyscan_t yyscanner) nogil

    void yypush_buffer_state(YY_BUFFER_STATE new_buffer, yyscan_t yyscanner) nogil
    void yypop_buffer_state(yyscan_t yyscanner) nogil
    YY_BUFFER_STATE yy_create_buffer(FILE* file, int size, yyscan_t yyscanner) nogil
    YY_BUFFER_STATE yy_scan_buffer(char* base, Py_ssize_t size, yyscan_t yyscanner) nogil
    YY_BUFFER_STATE yy_scan_bytes(const char* text, int len, yyscan_t yyscanner) nogil


from functools import partialmethod
from types import MethodType
from typing import Any, Callable, Dict, Generator, Iterable, Tuple, Union, overload
from typing_extensions import Self, Protocol, runtime_checkable


class CommandCaller(Protocol):
    def __call__(self, __command: Any, __args: tuple, __kwargs: Dict[str, Any], **kwds: Any) -> Any: ...


@runtime_checkable
class CommandLike(Protocol):
    def __kola_command__(self) -> Generator[Tuple[str, Callable], None, None]: ...


class Command(object):
    __slots__ = ["__name__", "__func__", "suppression", "virtual", "alias", "extra_data"]

    def __init__(
        self,
        __name: str,
        func: Callable,
        *,
        alias: Union[Iterable[str], str] = (),
        suppression: bool = False,
        virtual: bool = False,
        **kwds: Any
    ) -> None:
        self.__name__ = __name
        self.__func__ = func
        self.alias = (alias,) if isinstance(alias, str) else tuple(alias)
        self.suppression = suppression
        self.virtual = virtual
        self.extra_data = kwds
    
    @overload
    def set_data(self, __name: str, value: Any) -> Self: ...
    @overload
    def set_data(self, __name: str) -> Callable[[Any], Self]: ...

    def set_data(self, __name: str, value: Any = None) -> Union[Self, Callable[[Any], Self]]:
        if value is not None:
            self.extra_data[__name] = value
            return self
        
        def wrapper(val: Any) -> Self:
            self.extra_data[__name] = val
            return self
        return wrapper

    writer = partialmethod(set_data, "writer_func")
    
    @classmethod
    def from_command(cls, command: "Command", **kwds: Any) -> Self:
        data = command.extra_data.copy()
        data.update(kwds)
        return cls(
            command.__name__,
            command.__func__,
            alias=command.alias,
            **data
        )
    
    @property
    def __wrapped__(self) -> Callable:  # pragma: no cover
        return self.__func__
    
    def __kola_command__(self, force: bool = False) -> Generator[Tuple[str, Self], None, None]:
        if not self.suppression or force:
            bound_func = self
            yield self.__name__, bound_func
            for i in self.alias:
                yield i, bound_func

    def __get__(self, ins: Any, owner: type) -> Any:
        if ins is None:
            return self
        elif self.virtual and (not hasattr(owner, "check_virtual") or ins.check_virtual(self)):
            return ins[self.__name__]
        return MethodType(self, ins)

    def __call__(self, vmobj: Any, *args: Any, **kwds: Any) -> Any:
        caller = getattr(vmobj, "__kola_caller__", None)
        if caller is None:  # pragma: no cover
            return self.__func__(vmobj, *args, **kwds)
        return caller(self, args, kwds, **self.extra_data)

    def __repr__(self) -> str:
        return f"<kola command {self.__name__} with {self.__func__}>"



import sys
from types import TracebackType
from typing import Optional, Type
from traceback import print_exception

from kola import __version__
from kola.klvm import KoiLang, kola_command


class BaseDebugger(KoiLang):
    """
    provide unified version checking and error suppression functions
    """
    __slots__ = []

    @kola_command
    def version(self, __chk_ver: Optional[int] = None) -> None:
        if __chk_ver is None:
            print(__version__)
        elif __chk_ver < 100 or __chk_ver > 120:
            print(f"version {__chk_ver} is not support by runner {__version__}")
            sys.exit(2)

    def on_exception(self, exc_type: Type[BaseException], exc_ins: BaseException, traceback: TracebackType) -> bool:
        super().on_exception(exc_type, exc_ins, traceback)
        print_exception(exc_type, exc_ins, traceback)
        return True



from .command import Command, CommandLike
from .commandset import CommandSet, CommandSetMeta
from .environment import Environment
from .koilang import KoiLang
from .handler import AbstractHandler, default_handler
from .decorator import *


kola_env_class = kola_environment


__all__ = [
    "CommandLike",
    "Command",

    "CommandSetMeta",
    "CommandSet",
    "Environment",
    "KoiLang",
    
    "AbstractHandler",
    "default_handler",

    "kola_command",
    "kola_text",
    "kola_number",
    "kola_annotation",
    "kola_env_enter",
    "kola_env_exit",
    "kola_command_set",
    "kola_environment",
    "kola_main"
]



from abc import abstractmethod
from inspect import currentframe
from typing import Any, ClassVar, Dict, Optional, Tuple, Type, Union

from kola.klvm import Command, CommandSet, Environment, KoiLang, AbstractHandler
from kola.klvm.environment import EnvironmentAutopop


class EnvRunner(KoiLang):
    __slots__ = ["env_class"]

    def __init__(self, env_class: Type[Environment]) -> None:
        super().__init__()
        self.env_class = env_class
    
    def at_start(self) -> None:
        if self.top is self:
            self.push_apply(self.push_prepare(self.env_class))
    

class HnadlerEnv(Environment):
    __slots__ = ["_handler"]

    Handler: ClassVar = AbstractHandler

    def __init__(self, back: CommandSet) -> None:
        super(Environment, self).__init__()
        self.back = back

        if self.__class__.__env_autopop__:
            # for these have no exit point, use the same name as entry points
            for c in self.__class__.__env_entry__:
                self.raw_command_set.update(
                    EnvironmentAutopop.from_command(c).__kola_command__()
                )

    def set_up(self, top: CommandSet) -> None:
        self._handler = self.home.add_handler(self.__class__.Handler)
        super().set_up(top)
    
    def tear_down(self, top: CommandSet) -> None:
        super().tear_down(top)
        self.home.remove_handler(self._handler)
        del self._handler


class WrapperEnv(HnadlerEnv):
    __slots__ = []
    
    _handler: "WrapperEnv.Handler"

    @abstractmethod
    def wrapper(self, command: Command, args: Tuple, kwargs: Dict[str, Any], **kwds: Any) -> Any:
        return self.pass_down()
    
    def pass_down(
        self,
        command: Union[Command, None] = None,
        args: Optional[tuple] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        **kwds: Any
    ) -> Any:
        frame = currentframe()
        assert frame and frame.f_back, "failed to fetch frames"
        frame = frame.f_back
        try:
            assert frame.f_locals["self"] is self
            command = command or frame.f_locals["command"]
            args = args or frame.f_locals["args"]
            kwargs = kwargs or frame.f_locals["kwargs"]
            if not kwds:
                kwds.update(frame.f_locals["kwds"])
        except Exception as e:
            raise RuntimeError("failed to fetch arguments") from e
        handler_frame = frame.f_back
        assert handler_frame
        try:
            handler: WrapperEnv.Handler = handler_frame.f_locals["self"]
            assert handler.bound_env is self
        except Exception as e:
            raise RuntimeError("Incorrect handler object call stack frame") from e
        return super(handler.__class__, handler).__call__(command, args, kwargs, **kwds)
    
    def set_up(self, top: CommandSet) -> None:
        super().set_up(top)
        self._handler.bound(self)

    class Handler(AbstractHandler):
        __slots__ = ["bound_env"]

        priority = 1

        def bound(self, env: "WrapperEnv") -> None:
            self.bound_env = env

        def __call__(self, command: Command, args: Tuple, kwargs: Dict[str, Any], **kwds: Any) -> Any:
            return self.bound_env.wrapper(command, args, kwargs, **kwds)


env_runner = EnvRunner



import os
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple, Union, overload

from kola.lexer import BaseLexer, FileLexer, StringLexer
from kola.parser import Parser
from kola.klvm import CommandSet


class Instruction(NamedTuple):
    name: str
    args: Tuple[Any, ...]
    kwargs: Dict[Any, Any]


class _Recorder(CommandSet):
    __slots__ = []

    def __getitem__(self, __key: str) -> Callable[..., Instruction]:
        return lambda *args, **kwds: Instruction(__key, args, kwds)


recorder = _Recorder()


@overload
def parse(__lexer: BaseLexer) -> List[Instruction]: ...
@overload
def parse(__str: str, *, command_threshold: int = 1, no_lstrip: bool = ...) -> List[Instruction]: ...
def parse(__lexer_or_str: Union[str, BaseLexer], **kwds: Any) -> List[Instruction]:  # noqa: E302
    if isinstance(__lexer_or_str, str):
        __lexer_or_str = StringLexer(__lexer_or_str, **kwds)
    else:
        assert not kwds
    return list(Parser(__lexer_or_str, recorder))


def parse_file(path: Union[str, bytes, os.PathLike], *, encoding: Optional[str] = None, **kwds: Any) -> List[Instruction]:
    lexer = FileLexer(path, encoding=encoding or "utf-8", **kwds)
    return parse(lexer)



from functools import partial
from types import TracebackType
from typing import Any, Callable, Dict, Generator, Iterable, Optional, Set, Tuple, Type, TypeVar, Union, overload
from typing_extensions import Self

from ..exception import KoiLangError

from .command import Command
from .commandset import CommandSet, CommandSetMeta


T = TypeVar("T")


class EnvironmentCommand(Command):
    __slots__ = ["env_class"]

    def __init__(
        self,
        __name: str,
        func: Callable,
        env_class: Optional[Type["Environment"]] = None,
        envs: Union[Iterable[str], str, None] = None,
        **kwds
    ) -> None:
        super().__init__(__name, func, **kwds)
        if envs:
            self.extra_data["envs"] = (envs,) if isinstance(envs, str) else tuple(envs)
        self.env_class = env_class
    
    @classmethod
    def from_command(cls, command: "Command", **kwds: Any) -> Self:
        if isinstance(command, EnvironmentCommand):
            kwds["env_class"] = command.env_class
        return super().from_command(command, **kwds)


class EnvironmentEntry(EnvironmentCommand):
    __slots__ = []
    
    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, suppression=True, **kwds)
    
    def __get__(self, ins: Any, owner: type) -> Any:
        assert self.env_class
        if isinstance(ins, self.env_class) and self.env_class.__env_autopop__:
            return EnvironmentAutopop.from_command(self).__get__(ins, owner)
        return super().__get__(ins, owner)

    def __call__(self, vmobj: Union["Environment", "KoiLang"], *args: Any, **kwds: Any) -> Any:
        assert self.env_class
        home = vmobj.home
        env = home.push_prepare(self.env_class)
        ret = super().__call__(env, *args, **kwds)
        home.push_apply(env)
        return ret


class EnvironmentExit(EnvironmentCommand):
    __slots__ = []

    def __call__(self, vmobj: "Environment", *args: Any, **kwds: Any) -> Any:
        home = vmobj.home
        pop_env = home.pop_prepare(self.env_class)
        ret = super().__call__(vmobj, *args, **kwds)
        home.pop_apply(pop_env)
        return ret


class EnvironmentAutopop(EnvironmentCommand):
    __slots__ = []

    def __call__(self, vmobj: "Environment", *args: Any, **kwds: Any) -> Any:
        assert self.env_class
        home = vmobj.home

        cache = home.pop_prepare(self.env_class)
        home.pop_apply(cache)
        cache = home.push_prepare(self.env_class)
        ret = super().__call__(cache, *args, **kwds)
        home.push_apply(cache)
        return ret


class EnvironmentMeta(CommandSetMeta):
    __env_entry__: Set[EnvironmentCommand]
    __env_exit__: Set[EnvironmentCommand]

    def __new__(cls, name: str, bases: Tuple[type, ...], attr: Dict[str, Any], **kwds: Any) -> Self:
        entry = set()
        exit = set()
        
        # Usually, there are not many entry points and exit points.
        # Therefore, unlike the attribute `__command_fields__`, they are computed here
        for i in bases:
            if not isinstance(i, EnvironmentMeta):
                continue
            entry.update(i.__env_entry__)
            exit.update(i.__env_exit__)

        attr["__env_entry__"] = entry
        attr["__env_exit__"] = exit
        new_cls = super().__new__(cls, name, bases, attr, **kwds)
        
        for v in attr.values():
            if not isinstance(v, EnvironmentCommand):
                continue
            v.env_class = new_cls  # type: ignore
            if isinstance(v, EnvironmentEntry):
                entry.add(v)
            elif isinstance(v, EnvironmentExit):
                exit.add(v)
        return new_cls
    
    @property
    def __env_autopop__(self) -> bool:
        return not self.__env_exit__
    
    def __kola_command__(self) -> Generator[Tuple[str, EnvironmentCommand], None, None]:
        for i in self.__env_entry__:
            yield from i.__kola_command__(force=True)

    @overload
    def __get__(self, ins: Any, owner: type) -> Self: ...
    @overload  # noqa: E301
    def __get__(
        self: Type[T],
        ins: Union["KoiLang", "Environment"],
        owner: Union[Type["KoiLang"], Type["Environment"]]
    ) -> T: ...
    def __get__(self, ins: Any, owner: type) -> Any:  # noqa: E301
        if ins is not None and issubclass(owner, CommandSet):
            home: KoiLang = ins.home
            cur = home.top
            while isinstance(cur, Environment):
                if cur.back is ins and isinstance(cur, self):
                    return cur
                cur = cur.back
            else:
                if ins is home.top:
                    return self.EntryPointInterface(self, ins)
                raise ValueError(f"cannot find env '{self.__name__}' in the {home}")
        return self
    
    class EntryPointInterface:
        """
        interface for entry points in Python level
        """
        __slots__ = ["__env_class", "__bound_ins"]

        def __init__(self, env_class: "EnvironmentMeta", bound_ins: CommandSet) -> None:
            self.__env_class = env_class
            self.__bound_ins = bound_ins
        
        def __getattr__(self, __name: str) -> Any:
            attr = getattr(self.__env_class, __name)
            if isinstance(attr, EnvironmentEntry):
                return attr.__get__(self.__bound_ins, self.__bound_ins.__class__)
            elif isinstance(attr, Command):
                raise AttributeError(f"cannot fetch command '{__name}' before the environment initialization")
            raise AttributeError("only entry commands can be accessed through the interface")
        
        def __repr__(self) -> str:  # pragma: no cover
            return f"<kola environment '{self.__env_class.__name__}' entry point command interface>"


class Environment(CommandSet, metaclass=EnvironmentMeta):
    __slots__ = ["back"]

    def __init__(self, back: CommandSet) -> None:
        if not self.__class__.__env_entry__:
            raise TypeError(
                "cannot instantiate an environment that has no entry points"
            )
        super().__init__()
        self.back = back

        if self.__class__.__env_autopop__:
            # for these have no exit point, use the same name as entry points
            for c in self.__class__.__env_entry__:
                self.raw_command_set.update(
                    EnvironmentAutopop.from_command(c).__kola_command__()
                )

    def __getitem__(self, __key: str) -> Callable:
        cmd_set = self
        cmd = cmd_set.get(__key)
        while cmd is None:
            if not isinstance(cmd_set, Environment):
                raise KeyError(f"unknown command '{__key}'")
            cmd_set = cmd_set.back
            cmd = cmd_set.get(__key)
        return cmd
    
    def __kola_caller__(self, command: Command, args: tuple, kwargs: Dict[str, Any], **kwds: Any) -> Any:
        return self.home.__kola_caller__(command, args, kwargs, bound_instance=self, **kwds)

    @property
    def home(self) -> "KoiLang":
        cmd_set = self
        while isinstance(cmd_set, Environment):
            cmd_set = cmd_set.back
        assert isinstance(cmd_set, KoiLang)
        return cmd_set
    
    def set_up(self, top: CommandSet) -> None:
        """called before the environment added to the env stack top"""

    def tear_down(self, top: CommandSet) -> None:
        """called after the environment removed from the env stack top"""

    @partial(Command, "@end")
    def at_end(self) -> None:
        """
        ensure autopop environments properly popped at the end of parsing
        """
        if not self.__class__.__env_autopop__:
            return
        home = self.home
        home.pop_apply(home.pop_prepare(self.__class__))
        home.at_end()
    
    @partial(Command, "@exception", virtual=True, suppression=True)
    def on_exception(self, exc_type: Type[KoiLangError], exc_ins: Optional[KoiLangError], traceback: TracebackType) -> Any:
        return self.back["@exception"]


from .koilang import KoiLang



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Type, Union


from .commandset import CommandSet


class Mask(ABC):
    __slots__ = ["type", "not_"]

    class MType(Enum):
        default = ""    # test all reachable env
        exact = "+"     # test top env
        all = "*"      # test all env

    def __init__(self, type: Union[MType, str] = MType.default, not_: bool = False) -> None:
        if isinstance(type, str):
            type = self.MType(type)
        self.type = type
        self.not_ = not_

    @abstractmethod
    def __contains__(self, other: Any) -> bool:
        raise NotImplementedError


class ClassTypeMask(Mask):
    __slots__ = ["env_type"]

    def __init__(
        self,
        env_type: Type[CommandSet],
        type: Union[Mask.MType, str] = Mask.MType.default,
        not_: bool = False
    ) -> None:
        self.env_type = env_type
        super().__init__(type=type, not_=not_)
    
    def __contains__(self, other: Any) -> bool:
        return isinstance(other, self.env_type)


class ClassNameMask(Mask):
    __slots__ = ["env_name"]

    def __init__(
        self,
        env_name: str,
        type: Union[Mask.MType, str, None] = None,
        not_: bool = False,
        **kwds: CommandSet
    ) -> None:
        if env_name.startswith('!'):
            env_name = env_name[1:]
            not_ = True
        if type is None:
            if env_name.startswith('+'):
                env_name = env_name[1:]
                type = Mask.MType.exact
            elif env_name.startswith('*'):
                env_name = env_name[1:]
                type = Mask.MType.all
            else:
                type = Mask.MType.default
        elif env_name[0] in ('+', '*'):
            raise ValueError(
                "the name prefix should not be used to specify the type when the Mask type is specified"
            )
        if env_name.startswith('$'):
            name = env_name[1:]
            if name in kwds:
                env_name = kwds[name].__class__.__name__
        elif env_name == "__init__":
            env_name = kwds["__init__"].__class__.__name__
        self.env_name = env_name
        super().__init__(type=type, not_=not_)
    
    def __contains__(self, other: Any) -> bool:
        return self.env_name == other.__class__.__name__



"""
Kola builtin debugger module
"""

from kola.lib import KolaSpec, collect_all
from .base import BaseDebugger
from .command_debugger import CommandDebugger
from .default_runner import KoiLangRunner


__kola_spec__ = KolaSpec(
    "debugger",
    collect_all(),
    main_class=KoiLangRunner,
    doc=__doc__
)



"""
Copyright 2023 Ovizro

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import re
import sys
from types import ModuleType
from typing import Any, Dict, List, Optional, TypeVar, Union
from typing_extensions import Literal

from kola.exception import KoiLangCommandError
from kola.klvm import Environment, kola_command, kola_text
from kola.lib import KOLA_LIB_PATH, KolaSpec, collect_all, load_library, main_class_from_module
from kola.lib.debugger.base import BaseDebugger


T = TypeVar("T")


FLAG_DEBUG = 1


class KoiLangRunner(BaseDebugger):
    """
    default class for KoiLang module runner
    """

    var_pattern = re.compile(r"\$(\{)?(?P<name>[a-zA-Z_]\w*)(?(1)\}|)", re.ASCII)

    def __init__(self) -> None:
        super().__init__()
        self.flag = 0

    @kola_command
    def license(self) -> None:
        print(__doc__)
    
    @kola_command(alias="raise")
    def raises(self, *msg: str) -> None:
        raise KoiLangCommandError(*msg)
    
    @kola_command
    def pragma(self, *, debug: Optional[Literal["on", "off"]] = None) -> None:
        if debug:
            if debug == "on":
                self.flag |= FLAG_DEBUG
            elif debug == "off":
                self.flag &= ~FLAG_DEBUG
            else:
                raise ValueError(f"illegal value '{debug}'")
    
    @kola_command
    def echo(self, *text: str) -> None:
        print(' '.join(text))
    
    @kola_command
    def execute(self, source: str) -> None:
        code = compile(source, "<kola_runner>", "single")
        exec(code, {"kola_runner": self, "__vars__": self.vars})
    
    @kola_command(alias="export")
    def set(self, **kwds) -> None:
        self.vars.update(kwds)
    
    @kola_command
    def exit(self, code: int = 0) -> None:
        sys.exit(code)
        
    @kola_text
    def text(self, text: str) -> None:
        print(f"## [TEXT] {text}")
    
    @kola_command
    def load(self, path: Union[str, bytes, os.PathLike], type: Optional[str] = None, **kwds: Any) -> None:
        path = os.fsdecode(path)
        if type is None:
            if path.endswith(".kola"):
                type = "kola"
            else:
                type = "script"
        else:
            type = type.lower()
        if type == "kola":
            self.load_kola(path, **kwds)
        elif type in ["script", "lib"]:
            self.load_script(path, **kwds)
        else:
            raise ValueError(f"invalid type name '{type}'")
    
    @kola_command
    def load_kola(self, path: Union[str, bytes, os.PathLike], *, encoding: str = "utf-8", **kwds) -> None:
        self.parse_file(path, encoding=encoding, **kwds)
    
    @kola_command(alias=["load_lib", "load_library"])
    def load_script(self, name: str, lib_path: Optional[List[str]] = None) -> None:
        kola_module = load_library(name, lib_path or KOLA_LIB_PATH)
        self.load_command_set(kola_module)
    
    def at_start(self) -> None:
        self.vars = {}
        self.flag = 0
        self.loaded_class = set()
    
    def at_end(self) -> None:
        super().at_end()
        for i in self.loaded_class:
            i.at_end(self)
    
    def get_var(self, key: str) -> Any:
        if key == '__top__':
            return self.top.__class__.__name__
        elif key == "__dir__":
            return ', '.join(self.raw_command_set)
        elif key == "__name__":
            return self.__class__.__name__
        elif key == "__stack_info__":
            env_names = []
            top = self.top
            while top is not self:
                env_names.append(top.__class__.__name__)
                top = top.back  # type: ignore
            env_names.append("__init__")
            return " -> ".join(env_names)
        else:
            return self.vars.get(key, None)
    
    def load_command_set(self, module: ModuleType) -> None:
        cls = main_class_from_module(module)
        if self.flag & FLAG_DEBUG:
            print(f"## [DEBUG] Load class: {cls}")
        self.loaded_class.add(cls)
        cls.at_start(self)
        self.raw_command_set.update(
            cls.generate_raw_commands()
        )
    
    def format_text(self, argument: T) -> T:
        if isinstance(argument, list):
            return [self.format_text(i) for i in argument]  # type: ignore
        elif isinstance(argument, dict):
            return {k: self.format_text(v) for k, v in argument.items()}  # type: ignore
        elif isinstance(argument, str):
            return self.var_pattern.sub(
                lambda match: str(self.get_var(match.group("name")) or ''),
                argument
            )
        return argument
    
    def __kola_caller__(
            self, command: Any, args: tuple, kwargs: Dict[str, Any],
            *, bound_instance: Any = None, **kwds: Any) -> Any:
        if self.flag & FLAG_DEBUG:
            print(f"## [DEBUG] Run command: {command.__name__} ({(bound_instance or self).__class__.__name__})")
        return super().__kola_caller__(
            command,
            tuple(self.format_text(i) for i in args),
            self.format_text(kwargs),
            bound_instance=bound_instance,
            **kwds
        )
    
    def push_apply(self, __push_cache: Environment) -> None:
        if self.flag & FLAG_DEBUG:
            print(f"## [DEBUG] Push env: {__push_cache}")
        return super().push_apply(__push_cache)
    
    def pop_apply(self, __env_cache: Environment) -> None:
        if self.flag & FLAG_DEBUG:
            print(f"## [DEBUG] Pop env: {__env_cache}")
        return super().pop_apply(__env_cache)


__kola_spec__ = KolaSpec(
    "Kola Runner",
    collect_all(),
    main_class=KoiLangRunner
)



"""
simple python module for KoiLang parsing
"""

from .lexer import BaseLexer, FileLexer, StringLexer
from .parser import Parser
from .writer import BaseWriter, FileWriter, StringWriter, BaseWriterItem, FormatItem, ComplexArg, WriterItemLike
from .klvm import KoiLang, Environment, kola_command, kola_text, kola_number, kola_annotation, kola_env_enter, kola_env_exit, kola_env_class
from .version import __version__, __version_num__
from .exception import KoiLangError, KoiLangSyntaxError, KoiLangCommandError


__author__ = "Ovizro"
__author_email__ = "Ovizro@visecy.top"

__all__ = [
    "KoiLang",
    "Environment",
    "kola_command",
    "kola_text",
    "kola_number",
    "kola_annotation",
    "kola_env_enter",
    "kola_env_exit",
    
    "BaseLexer",
    "FileLexer",
    "StringLexer",
    "Parser",
    "BaseWriter",
    "FileWriter",
    "StringWriter",
    "BaseWriterItem",
    "FormatItem",
    "ComplexArg",
    "WriterItemLike",

    "KoiLangError",
    "KoiLangSyntaxError",
    "KoiLangCommandError"
]
